""" calculating molecular enhancment of collision.

    mol_enh =
        ( (vapor_size + particle_size) / (particle_size) ) ** 2

"""

import numpy as np
from particula.util.input_handling import in_length


def mol_enh(vapor_size, particle_size):
    """ Returns the molecular enhancement.

        Parameters:
            vapor_size      (float)  [m]
            particle_size   (float)  [m]

        Returns:
                    (float)  [ ]

        TODO:
            - Add explanation of molecular enhancement.
    """

    a_q = in_length(vapor_size)
    b_q = in_length(particle_size)

    return np.transpose(
        (np.transpose([a_q.m])*a_q.u + b_q) / (b_q)
    )**2



""" handling inputs
"""

from particula import u


def in_radius(radius):
    """ Handles radius input

        Parameters:
            radius    (float) [m | dimensionless]

        Returns:
            radius    (float) [m]

        Notes:
            * If unit is correct, take to base units in m
            * Throws ValueError if unit is wrong
            * Assigning m units to scalar input
    """

    if radius is None:
        raise ValueError("You must provide a radius!")

    if isinstance(radius, u.Quantity):
        if radius.to_base_units().u == u.m:
            radius = radius.to_base_units()
        else:
            raise ValueError(
                f"\n\t"
                f"Input {radius} has unsupported units.\n\t"
                f"Input must have radius units of\n\t"
                f"'meters';\n\t"
                f"otherwise, if dimensionless, it will\n\t"
                f"be assigned a m unit.\n"
            )
    else:
        radius = u.Quantity(radius, u.m)

    return radius


def in_handling(value, units: u.Quantity):
    """ generic function to handle inputs

        Parameters:
            value     (float)       [u.Quantity | dimensionless]
            units     (u.Quantity)

        Returns:
            value     (float)       [u.Quantity]

        Notes:
            * If unit is correct, take to base units
            * Throws ValueError if unit is wrong
            * Assigning default base units to scalar input
    """

    if isinstance(value, u.Quantity):
        if value.to_base_units().u == (1*units).to_base_units().u:
            value = value.to_base_units()
        else:
            raise ValueError(
                f"\n\t"
                f"Input has unsupported units.\n\t"
                f"Input must have units equivlanet to {units};\n\t"
                f"otherwise, if dimensionless, it will\n\t"
                f"be assigned {units}.\n"
            )
    else:
        value = u.Quantity(value, (1*units).to_base_units().u)

    return value


# pylint: disable=missing-docstring, multiple-statements
def in_temperature(temp): return in_handling(temp, u.K)
def in_viscosity(vis): return in_handling(vis, u.kg/u.m/u.s)
def in_pressure(pres): return in_handling(pres, u.Pa)
def in_mass(mass): return in_handling(mass, u.kg)
def in_volume(vol): return in_handling(vol, u.m**3)
def in_time(time): return in_handling(time, u.s)
def in_velocity(vel): return in_handling(vel, u.m/u.s)
def in_acceleration(acc): return in_handling(acc, u.m/u.s**2)
def in_molecular_weight(molw): return in_handling(molw, u.kg/u.mol)
def in_density(density): return in_handling(density, u.kg/u.m**3)
def in_scalar(scalar): return in_handling(scalar, u.dimensionless)
def in_length(length): return in_handling(length, u.m)
def in_area(area): return in_handling(area, u.m**2)
def in_gas_constant(con): return in_handling(con, u.J/u.K/u.mol)
def in_concentration(conc): return in_handling(conc, u.kg/u.m**3)



""" dimensionless coagulation according to several approx.
"""

import numpy as np
from particula.util.input_handling import in_scalar


def approx_coag_less(  # pylint: disable=too-many-locals
    diff_knu=None,
    cpr=None,
    approx="hardsphere"
):
    """ dy2007 approx.
        Dimensionless particle--particle coagulation kernel.

        gh2012:
        https://journals.aps.org/pre/abstract/10.1103/PhysRevE.85.026410

        gk2008:
        https://journals.aps.org/pre/abstract/10.1103/PhysRevE.78.046402

        dy2007:
        https://aip.scitation.org/doi/10.1063/1.2713719

        cg2019:
        https://www.tandfonline.com/doi/suppl/10.1080/02786826.2019.1614522

        todolater:
        - quick fixes for corner cases
        - examine better solutions?
    """

    if diff_knu is None:
        raise ValueError(
            "Please provide explicit value for diff_knu")
    if approx != "hardsphere":
        if cpr is None:
            raise ValueError("Please provide explicit value for cpr")
        cpr = in_scalar(cpr)
        cpr = cpr + 1e-16 if cpr.m == 0 else cpr  # avoid division by zero

        cekl = 1 + cpr
        cecl = 1 * cpr.u if cpr.m == 0 else cpr / (1 - np.exp(-cpr))

    diff_knu = in_scalar(diff_knu)

    coag_clim = 4 * np.pi * diff_knu**2

    hsa_consts = [25.836, 11.211, 3.502, 7.211]

    upstairs = (
        coag_clim +
        (hsa_consts[0] * diff_knu**3) +
        ((8 * np.pi)**(1/2) * hsa_consts[1] * diff_knu**4)
    )

    downstairs = (
        1 +
        (hsa_consts[2] * diff_knu) +
        (hsa_consts[3] * diff_knu**2) +
        (hsa_consts[1] * diff_knu**3)
    )

    hscoag = upstairs / downstairs

    if approx == "hardsphere":

        return hscoag

    if approx == "gh2012":

        min_fxn = np.min([diff_knu.m, 3*diff_knu.m/(2*cpr.m)])
        result = (4 * np.pi * diff_knu**2) / (
            1 + 1.598 * min_fxn**1.1709
        ) if cpr.m > 0.5 and min_fxn < 2.5 else hscoag

        return result

    if approx == "gk2008":

        result = coag_clim * (
            1 -
            (1 + (np.sqrt(np.pi)*cecl*cpr*1.22) / (
                2*cekl*diff_knu)) *
            np.exp(-(np.sqrt(np.pi)*cecl*cpr*1.22) / (
                2*cekl*diff_knu))
        ) + (np.sqrt(8*np.pi)*diff_knu) * (
            0 +
            (1 + (2*np.sqrt(np.pi)*(1.22**3)*cecl*(cpr**3)) / (
                9*(cekl**2)*diff_knu)) *
            np.exp(-(np.sqrt(np.pi)*cecl*cpr*1.22) / (
                2*cekl*diff_knu))
        )

        return (
            result.m * (cpr.m >= 0) + hscoag.m * (cpr.m < 0)
        ) * result.u

    if approx == "dy2007":

        return coag_clim / (
            np.sqrt(2*np.pi)*diff_knu*cekl *
            np.exp(-1*cpr/(1+diff_knu*cekl/cecl)) /
            (
                (1+diff_knu*cekl/cecl)**2 -
                (2+diff_knu*cekl/cecl)*diff_knu*cekl/cecl *
                np.exp(-1*cpr/(
                    (1+diff_knu*cekl/cecl) *
                    (2+diff_knu*cekl/cecl)
                ))
            ) +
            (1 - np.exp(-1*cpr/(1+diff_knu*cekl/cecl))) /
            (1 - np.exp(-1*cpr))
        )

    if approx == "cg2019":
        # fix later: 1e-12 is arbitrary, pure numerics issue here
        diff_knu = diff_knu if diff_knu.m > 1e-12 else 1e-12

        tricky_corr = 1 if cpr.m <= 0 else (
            4.528*np.exp(-1.088*cpr)) + (.7091*np.log(1+1.527*cpr))

        tricky_corr_again = 0 if cpr.m <= 0 else (
            11.36*(cpr**0.272) - 10.33)

        corr = [
            2.5,
            tricky_corr,
            tricky_corr_again,
            -0.003533*cpr + 0.05971
        ]

        corr_mu = (corr[2]/corr[0])*(
            (1+corr[3]*(np.log(diff_knu)-corr[1])/corr[0])**(
                -1/corr[3]-1) *
            np.exp(-1*(1+corr[3]*(np.log(diff_knu)-corr[1])/corr[0])**(
                -1/corr[3]))
            )

        return (
            hscoag.m * (cpr.m <= 0) + np.exp(corr_mu.m) * hscoag.m * (cpr > 0)
        ) * hscoag.u

    return None



""" calculating the particle area
"""

import numpy as np
from particula.util.input_handling import in_radius, in_scalar


def area(
    radius=None,
    area_factor=1,
):
    """ Returns particle's surface area: 4 pi r^2 .

        Parameters:
            radius       (float) [m]
            area_factor  (float) [ ]      (default: 1)

        Returns:
                         (float) [m^2]
    """

    radius = in_radius(radius)
    area_factor = in_scalar(area_factor)

    return 4*np.pi*(radius**2)*area_factor



""" determing the radius cutoff for particle distribution
"""

import numpy as np
from scipy.stats import lognorm

from particula.util.input_handling import in_scalar, in_radius


def cut_rad(
    cutoff=in_scalar(0.9999).m,
    gsigma=in_scalar(1.25).m,
    mode=in_radius(100e-9).m,
    **kwargs
):
    """ This routine determins the radius cutoff for the particle distribution

        Inputs:
            cutoff  (float) coverage cutoff (default: .9999)
            gsigma  (float) geometric standard deviation (default: 1.25)
            mode    (float) mean radius of the particles (default: 1e-7)

        Returns:
            (starting radius, ending radius) float tuple
    """

    _ = kwargs.get("something", None)

    if np.array([mode]).size == 1:
        (rad_start, rad_end) = lognorm.interval(
            alpha=cutoff,
            s=np.log(gsigma),
            scale=mode,
        )
    else:
        (rad_start_pre, rad_end_pre) = lognorm.interval(
            alpha=cutoff,
            s=np.log(gsigma),
            scale=mode,
        )
        rad_start = rad_start_pre.min()
        rad_end = rad_end_pre.max()

    return (rad_start, rad_end)



""" calculating Knudsen number

    TODO:
        add a sanity check for mixing radius shapes and temperature shapes
        for example, someone may want to calculate knu for [1e-9, 2e-9, 3e-9]
        but only at two different temperatures [290, 300].
        A solution is to broadcast into a new shape, e.g. (3, 2) with the
        leading dim being the radius; or we could simply ban this entirely.

        -- implementing tranpose for now.
"""

import numpy as np
from particula.util.input_handling import in_length, in_radius
from particula.util.mean_free_path import mfp as mfp_func


def knu(
    radius=None,
    mfp=None,
    **kwargs
):
    """ Returns particle's Knudsen number.

        The Knudsen number reflects the relative length scales of
        the particle and the suspending fluid (air, water, etc.).
        This is calculated by the mean free path of the medium
        divided by the particle radius.

        The Knudsen number is a measure of continuum effects and
        deviation thereof. For larger particles, the Knudsen number
        goes towards 0. For smaller particles, the Knudsen number
        goes towards infinity.

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.knudsen_number import knu
        >>> # with radius 1e-9 m
        >>> knu(radius=1e-9)
        <Quantity(66.4798498, 'dimensionless')>
        >>> # with radius 1e-9 m and mfp 60 nm
        >>> knu(radius=1e-9*u.m, mfp=60*u.nm).m
        60.00000000000001
        >>> calculating via mfp(**kwargs)
        >>> knu(
        ... radius=1e-9*u.m,
        ... temperature=300,
        ... pressure=1e5,
        ... molecular_weight=0.03,
        ... )
        <Quantity(66.7097062, 'dimensionless')>
        ```

        Parameters:
            radius  (float) [m]
            mfp     (float) [m] (default: util)

        Returns:
                    (float) [dimensionless]

        Notes:
            mfp can be calculated using mfp(**kwargs);
            refer to particula.util.mean_free_path.mfp for more info.

    """

    mfp_val = mfp_func(**kwargs) if mfp is None else in_length(mfp)
    radius = in_radius(radius)

    return np.transpose([mfp_val.m])*mfp_val.u / radius



""" a simple, fast, and powerful particle simulator.

    particula is a simple, fast, and powerful particle simulator,
    or at least two of the three, we hope. It is a simple particle
    system that is designed to be easy to use and easy to extend.
    The goal is to have a robust aerosol (gas phase + particle phase)
    simulation system that can be used to answer scientific questions
    that arise for experiments and research discussions.

    The main features of particula are:
    ...

    More details to follow.
"""

# Import pint here to avoid using a different registry in each module.
from pint import UnitRegistry

# u is the unit registry name.
u = UnitRegistry()

__version__ = "0.0.5"



""" a centralized location for important, unchanged physics parameters.

    This file contains constants that are used in multiple modules. Each
    constant has its own units and exported with them. The constants are
    mostly related to atmospheric aerosol particles in usual conditions.

"""

from particula import u

BOLTZMANN_CONSTANT = 1.380649e-23 * u.m**2 * u.kg / u.s**2 / u.K

AVOGADRO_NUMBER = 6.022140857e23 / u.mol

# Gas constant in J mol^-1 K^-1 = m^2 kg mol^-1 s^-2 K^-1
# J = kg m^2 s^-2
GAS_CONSTANT = BOLTZMANN_CONSTANT * AVOGADRO_NUMBER

ELEMENTARY_CHARGE_VALUE = 1.60217662e-19 * u.C

# Relative permittivity of air at approx.
# 296.15 K and 101325 Pa and 40% RH
# See https://www.osti.gov/servlets/purl/1504063
# Previously known as the "dielectric constant"
# Often denoted as epsilon
RELATIVE_PERMITTIVITY_AIR_ROOM = 1.000530569  # unitless
# At STP (273.15 K, 1 atm):
# see: https://en.wikipedia.org/wiki/Relative_permittivity
RELATIVE_PERMITTIVITY_AIR_STP = 1.00058986  # unitless

# select one of the two:
RELATIVE_PERMITTIVITY_AIR = RELATIVE_PERMITTIVITY_AIR_ROOM

# Permittivity of free space in F/m
# Also known as the electric constant, permittivity of free space
# Often denoted by epsilon_0
VACUUM_PERMITTIVITY = 8.85418782e-12 * u.F / u.m

ELECTRIC_PERMITTIVITY = RELATIVE_PERMITTIVITY_AIR * VACUUM_PERMITTIVITY

# These values are used to calculate the dynamic viscosity of air
# Here, REF temperature and viscosity are at STP:
# Standard temperature and pressure (273.15 K and 101325 Pa)
REF_VISCOSITY_AIR_STP = 1.716e-5 * u.Pa * u.s
REF_TEMPERATURE_STP = 273.15 * u.K
SUTHERLAND_CONSTANT = 110.4 * u.K
MOLECULAR_WEIGHT_AIR = (28.9644 * u.g / u.mol).to_base_units()



""" calculating the particle mass
"""

import numpy as np
from particula.util.input_handling import in_density, in_radius, in_scalar


def mass(
    radius=None,
    density=1000,
    shape_factor=1,
    volume_void=0,
    **kwargs,
):
    """ Returns particle's mass: 4/3 pi r^3 * density.

        Examples:
        ```
        >>> particle_mass(1*u.m)
        <Quantity(4188.7902, 'kilogram')>
        >>> particle_mass(1*u.nm, 1000*u.kg/u.m**3).m
        4.188790204786392e-24
        >>> particle_mass(1*u.nm, 1000*u.kg/u.m**3).m_as(u.g)
        4.188790204786392e-21
        >>> particle_mass([1, 2, 3]*u.nm, 1000*u.kg/u.m**3).m
        array([4.18879020e-24, 3.35103216e-23, 1.13097336e-22])
        >>> particle_mass([1, 2]*u.nm, [1, 2]*u.g/u.cm**3).m
        array([4.18879020e-24, 6.70206433e-23])
        >>> particle_mass(2*u.nm, 2*u.g/u.cm**3).m
        6.702064327658225e-23
        ```

        Parameters:
            radius       (float) [m]
            density      (float) [kg/m^3] (default: 1000)
            shape_factor (float) [ ]      (default: 1)
            volume_void  (float) [ ]      (default: 0)

        Returns:
                         (float) [kg]
    """
    _ = kwargs.get("something", None)
    radius = in_radius(radius)
    density = in_density(density)
    shape_factor = in_scalar(shape_factor)
    vol_void = in_scalar(volume_void)

    return (
        density * (4*np.pi/3) * (radius**3)
        * shape_factor * (1 - vol_void)
    ).to_base_units()



""" calculate the vapor flux to the surface of particles
"""

import numpy as np
from particula import u
from particula.util.fuchs_sutugin import fsc as fsc_func
from particula.util.input_handling import (in_concentration, in_handling,
                                           in_scalar)
from particula.util.molecular_enhancement import mol_enh
from particula.util.particle_surface import area
from particula.util.rms_speed import cbar


def phi(  # pylint: disable=too-many-arguments
    particle_area=None,
    molecular_enhancement=None,
    vapor_attachment=1,
    vapor_speed=None,
    driving_force=1,
    fsc=None,
    **kwargs,
):
    """ vapor flux
    """
    particle_area = area(
        **kwargs) if particle_area is None else in_handling(
            particle_area, u.m**2)
    molecular_enhancement_val = mol_enh(
        **kwargs) if molecular_enhancement is None else in_scalar(
            molecular_enhancement)
    vapor_attachment = in_scalar(
        np.array(
            [vapor_attachment.m]
            )*vapor_attachment.u)
    vapor_speed_val = cbar(
        **kwargs)/4 if vapor_speed is None else in_handling(
            vapor_speed, u.m/u.s)
    driving_force = in_concentration(driving_force)
    fsc_val = fsc_func(**kwargs) if fsc is None else in_scalar(fsc)

    result = (
        np.transpose([particle_area.m])*particle_area.u *
        np.transpose(
            [molecular_enhancement_val.m]
            )*molecular_enhancement_val.u *
        np.transpose([vapor_attachment.m])*vapor_attachment.u *
        np.transpose([vapor_speed_val.m])*vapor_speed_val.u *
        np.transpose([driving_force.m])*driving_force.u *
        np.transpose([fsc_val.m])*fsc_val.u
    ).squeeze()

    return np.transpose(result.m)*result.u



""" a class to initiate vapors condensing onto particles
"""
import numpy as np

from particula import u
from particula.environment import Environment
from particula.util.input_handling import (in_concentration, in_density,
                                           in_length, in_molecular_weight,
                                           in_scalar)


class Vapor(Environment):
    """ based on the Environment class
    """

    def __init__(self, **kwargs):
        """ initiating the vapor class
        """
        super().__init__(**kwargs)

        self.vapor_radius = in_length(
            kwargs.get('vapor_radius', 1.6e-9)
        )
        self.vapor_density = in_density(
            kwargs.get('vapor_density', 1400)
        )
        self.vapor_concentration = in_concentration(
            kwargs.get('vapor_concentration', 1)
        )
        self.vapor_attachment = in_scalar(
            kwargs.get('vapor_attachment', 1)
        )
        self.vapor_molec_wt = in_molecular_weight(
            kwargs.get('vapor_molecular_weight', 200*u.g/u.mol)
        )

        self.kwargs = kwargs

    def driving_force(self):
        """ condensation driving force
        """
        return np.array(
            [self.vapor_concentration.m]
        )*self.vapor_concentration.u



""" Calculate friction factor
"""

import numpy as np
from particula.util.dynamic_viscosity import dyn_vis
from particula.util.input_handling import in_radius, in_scalar, in_viscosity
from particula.util.slip_correction import scf


def frifac(
    radius=None,
    dynamic_viscosity=None,
    scf_val=None,
    **kwargs
):
    """ Returns a particle's friction factor.

        Property of the particle's size and surrounding medium.
        Multiplying the friction factor by the fluid velocity
        yields the drag force on the particle.

        It is best thought of as an inverse of mobility or
        the ratio between thermal energy and diffusion coefficient.
        The modified Stoke's diffusion coefficient is defined as
        kT / (6 * np.pi * dyn_vis_air * radius / slip_corr)
        and thus the friction factor can be defined as
        (6 * np.pi * dyn_vis_air * radius / slip_corr).

        In the continuum limit (Kn -> 0; Cc -> 1):
            6 * np.pi * dyn_vis_air * radius

        In the kinetic limit (Kn -> inf):
            8.39 * (dyn_vis_air/mfp_air) * const * radius**2

        See more: DOI: 10.1080/02786826.2012.690543 (const=1.36)

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.friction_factor import frifac
        >>> # with 1e-9 m radius
        >>> frifac(radius=1e-9)
        <Quantity(3.12763919e-15, 'kilogram / second')>
        >>> # with 1e-9 m radius and 1e-5 N/m^2 dynamic viscosity
        >>> frifac(radius=1e-9, dynamic_viscosity=1e-5)
        <Quantity(3.114213e-15, 'kilogram / second')>
        >>> # using dyn_vis(**kwargs)
        >>> frifac(
        ... radius=1e-9,
        ... temperature=298.15,
        ... reference_viscosity=1.716e-5,
        ... reference_temperature=273.15
        )
        <Quantity(3.12763919e-15, 'kilogram / second')>
        >>> # overriding sfc(**kwargs)
        >>> frifac(radius=1e-9, slip_correction=1.5)
        <Quantity(3.12763919e-15, 'kilogram / second')>
        ```
        Parameters:
            radius            (float) [m]
            dynamic_viscosity (float) [kg/m/s]  (default: util)
            slip_corr_factor  (float) [ ]       (default: util)

        Returns:
                              (float) [N*s/m]

        Notes:
            dynamic_viscosity can be calculated using the utility
            function particula.util.dynamic_viscosity.dyn_vis(**kwargs)
            and slip_corr_factor can be calculated using the utility
            function particula.util.slip_correction.scf(**kwargs);
            see respective documentation for more information.
    """

    radius = in_radius(radius)

    if dynamic_viscosity is None:
        dyn_vis_val = dyn_vis(radius=radius, **kwargs)
    else:
        dyn_vis_val = in_viscosity(dynamic_viscosity)

    if scf_val is None:
        scf_val = scf(radius=radius, **kwargs)
    else:
        scf_val = in_scalar(scf_val)

    return (
        6 *
        np.pi *
        np.transpose([dyn_vis_val.m])*dyn_vis_val.u *
        radius /
        scf_val
    ).to_base_units()



""" defining the environment class

    kwargs:
        temperature             (float) [K]      (default: 298.15)
        pressure                (float) [Pa]     (default: 101325)
      # dynamic_viscosity       (float) [Pa*s]   (default: util)
      $ molecular_weight        (float) [kg/mol] (default: constants)
      $ reference_viscosity     (float) [Pa*s]   (default: constants)
      $ reference_temperature   (float) [K]      (default: constants)
      $ sutherland_constant     (float) [K]      (default: constants)
      $ gas_constant            (float) [J/mol/K](default: constants)

      Using particula.util:
        dynamic_viscosity       (float) [Pa*s]

      Using particula.constants:
        GAS_CONSTANT            (float) [J/mol/K]
        MOLECULAR_WEIGHT_AIR    (float) [kg/mol]
        REF_VISCOSITY_AIR_STP   (float) [Pa*s]
        REF_TEMPERATURE_STP     (float) [K]
        SUTHERLAND_CONSTANT     (float) [K]

    Notes:
      # stands for an optional util override
      $ stands for an optional constants override

"""

from particula.constants import (GAS_CONSTANT, MOLECULAR_WEIGHT_AIR,
                                 REF_TEMPERATURE_STP, REF_VISCOSITY_AIR_STP,
                                 SUTHERLAND_CONSTANT)
from particula.util.dynamic_viscosity import dyn_vis
from particula.util.input_handling import (in_gas_constant,
                                           in_molecular_weight, in_pressure,
                                           in_temperature, in_viscosity)
from particula.util.mean_free_path import mfp


class Environment:  # pylint: disable=too-many-instance-attributes
    """ creating the environment class

        For now, the environment class takes properties such as
        temperature and pressure to calculate derived properties
        such as viscosity and mean free path.
    """

    def __init__(self, **kwargs):
        """ Initiate the environment class with base attrs.
        """

        self.temperature = in_temperature(
            kwargs.get("temperature", 298.15)
        )
        self.reference_viscosity = in_viscosity(
            kwargs.get("reference_viscosity", REF_VISCOSITY_AIR_STP)
        )
        self.reference_temperature = in_temperature(
            kwargs.get("reference_temperature", REF_TEMPERATURE_STP)
        )
        self.pressure = in_pressure(
            kwargs.get("pressure", 101325)
        )
        self.molecular_weight = in_molecular_weight(
            kwargs.get("molecular_weight", MOLECULAR_WEIGHT_AIR)
        )
        self.sutherland_constant = in_temperature(
            kwargs.get("sutherland_constant", SUTHERLAND_CONSTANT)
        )
        self.gas_constant = in_gas_constant(
            kwargs.get("gas_constant", GAS_CONSTANT)
        )

        self.kwargs = kwargs

    def dynamic_viscosity(self):
        """ Returns the dynamic viscosity in Pa*s.
        """
        return dyn_vis(
            temperature=self.temperature,
            reference_viscosity=self.reference_viscosity,
            reference_temperature=self.reference_temperature,
            sutherland_constant=self.sutherland_constant,
        )

    def mean_free_path(self):
        """ Returns the mean free path in m.
        """
        return mfp(
            temperature=self.temperature,
            pressure=self.pressure,
            molecular_weight=self.molecular_weight,
            dynamic_viscosity=self.dynamic_viscosity(),
            gas_constant=self.gas_constant,
        )



""" calculating the dynamic viscosity

    The dynamic viscocity is calculated using the Sutherland formula,
    assuming ideal gas behavior, as a function of temperature.

        "The dynamic viscosity equals the product of the sum of
        Sutherland's constant and the reference temperature divided by
        the sum of Sutherland's constant and the temperature,
        the reference viscosity and the ratio to the 3/2 power
        of the temperature to reference temperature."

        https://resources.wolframcloud.com/FormulaRepository/resources/Sutherlands-Formula

"""
from particula import u
from particula.constants import (REF_TEMPERATURE_STP, REF_VISCOSITY_AIR_STP,
                                 SUTHERLAND_CONSTANT)
from particula.util.input_handling import in_temperature, in_viscosity


def dyn_vis(
    temperature=298.15 * u.K,
    reference_viscosity=REF_VISCOSITY_AIR_STP,
    reference_temperature=REF_TEMPERATURE_STP,
    sutherland_constant=SUTHERLAND_CONSTANT,
    **kwargs,
):
    """ The dynamic viscosity of air via Sutherland formula.
        This formula depends on temperature (temp) and the reference
        temperature (t_ref) as well as the reference viscosity (mu_ref).

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.dynamic_viscosity import dyn_vis
        >>> # with units
        >>> dyn_vis(
        ... temperature=298.15*u.K,
        ... reference_viscosity=1.716e-5*u.Pa*u.s
        ... )
        <Quantity(1.83714937e-05, 'kilogram / meter / second')>
        >>> # without units and taking magnitude
        >>> dyn_vis(
        ... temperature=298.15,
        ... reference_viscosity=1.716e-5
        ... ).magnitude
        1.8371493734583912e-05
        >>> # without units, all keyword arguments
        >>> dyn_vis(
        ... temperature=298.15,
        ... reference_viscosity=1.716e-5,
        ... reference_temperature=273.15
        ... )
        <Quantity(1.83714937e-05, 'kilogram / meter / second')>
        >>> # for a list of temperatures
        >>> dyn_vis(temperature=[200, 250, 300, 400]).m
        array([1.32849751e-05, 1.59905239e-05, 1.84591625e-05, 2.28516090e-05])
        ```

        Inputs:
            temperature             (float) [K]     (default: 298.15)
            reference_viscosity     (float) [Pa*s]  (default: constants)
            reference_temperature   (float) [K]     (default: constants)
            sutherland_constant     (float) [K]     (default: constants)

        Returns:
                                    (float) [Pa*s]

        Using particula.constants:
            REF_VISCOSITY_AIR_STP   (float) [Pa*s]
            REF_TEMPERATURE_STP     (float) [K]
            SUTHERLAND_CONSTANT     (float) [K]

    """
    # trick to avoid triggering a linting error for kwargs
    temp = kwargs.get("temperature", temperature)
    temp = in_temperature(temperature)
    ref_vis = in_viscosity(reference_viscosity)
    ref_temp = in_temperature(reference_temperature)
    suth_const = in_temperature(sutherland_constant)

    return (
        ref_vis *
        (temp/ref_temp)**(3/2) *
        (ref_temp + suth_const) /
        (temp + suth_const)
    ).to_base_units()



""" the Fuchs-Sutugin model transition regime correction
"""

import numpy as np
from particula.util.knudsen_number import knu
from particula.util.input_handling import in_scalar


def fsc(
    knu_val=None,
    alpha=1,
    **kwargs
):
    """ Returns the Fuchs-Sutugin model transition regime correction.

        Parameters:
            knu     (float)  [ ] (default: util)
            alpha   (float)  [ ] (default: 1)

        Returns:
                (float)  [ ]

        Notes:
            knu can be calculated using knu(**kwargs);
            refer to particula.util.knudsen_number.knu for more info.
    """

    knu_val = knu(**kwargs) if knu_val is None else in_scalar(knu_val)
    alpha = np.transpose([in_scalar(alpha).m])*in_scalar(alpha).u

    return np.transpose(
        knu_val * alpha * (1 + knu_val) /
        (knu_val**2 + knu_val + 0.283 * knu_val * alpha + 0.75 * alpha)
    )



""" the particule class
"""

import numpy as np

from particula import u
from particula.constants import (AVOGADRO_NUMBER, BOLTZMANN_CONSTANT,
                                 ELECTRIC_PERMITTIVITY,
                                 ELEMENTARY_CHARGE_VALUE)
from particula.util.dimensionless_coagulation import DimensionlessCoagulation
from particula.util.distribution_discretization import discretize
from particula.util.friction_factor import frifac
from particula.util.fuchs_sutugin import fsc
from particula.util.input_handling import (in_density, in_handling, in_radius,
                                           in_scalar, in_volume)
from particula.util.knudsen_number import knu
from particula.util.molecular_enhancement import mol_enh
from particula.util.particle_mass import mass
from particula.util.particle_surface import area
from particula.util.radius_cutoff import cut_rad
from particula.util.reduced_quantity import reduced_quantity as redq
from particula.util.rms_speed import cbar
from particula.util.slip_correction import scf
from particula.util.vapor_flux import phi
from particula.vapor import Vapor


class ParticleDistribution(Vapor):
    """ starting a particle distribution from continuous pdf
    """

    def __init__(self, **kwargs):
        """  particle distribution objects.
        """
        super().__init__(**kwargs)

        self.spacing = kwargs.get("spacing", "linspace")
        self.nbins = in_scalar(kwargs.get("nbins", 1000)).m
        self.nparticles = in_scalar(kwargs.get("nparticles", 1e5)).m
        self.volume = in_volume(kwargs.get("volume", 1e-6))
        self.cutoff = in_scalar(kwargs.get("cutoff", 0.9999)).m
        self.gsigma = in_scalar(kwargs.get("gsigma", 1.25)).m
        self.mode = in_radius(kwargs.get("mode", 100e-9)).m

    def pre_radius(self):
        """ Returns the radius space of the particles

            Utilizing the utility cut_rad to get 99.99% of the distribution.
            From this interval, radius is made on a linspace with nbins points.
            Note: linspace is used here to practical purposes --- often, the
            logspace treatment will return errors in the discretization due
            to the asymmetry across the interval (finer resolution for smaller
            particles, but much coarser resolution for larger particles).
        """

        (rad_start, rad_end) = cut_rad(
            cutoff=self.cutoff,
            gsigma=self.gsigma,
            mode=self.mode,
        )

        if self.spacing == "logspace":
            radius = np.logspace(
                np.log10(rad_start),
                np.log10(rad_end),
                np.array([self.nbins]).sum()
            )
        elif self.spacing == "linspace":
            radius = np.linspace(
                rad_start,
                rad_end,
                np.array([self.nbins]).sum()
            )
        else:
            raise ValueError("Spacing must be 'logspace' or 'linspace'!")

        return radius*u.m

    def pre_discretize(self):
        """ Returns a distribution pdf of the particles

            Utilizing the utility discretize to get make a lognorm distribution
            via scipy.stats.lognorm.pdf:
                interval: the size interval of the distribution
                gsigma  : geometric standard deviation of distribution
                mode    : geometric mean radius of the particles
        """

        return discretize(
            interval=self.pre_radius(),
            disttype="lognormal",
            gsigma=self.gsigma,
            mode=self.mode,
            nparticles=self.nparticles
        )

    def pre_distribution(self):
        """ Returns a distribution pdf of the particles

            Utilizing the utility discretize to get make a lognorm distribution
            via scipy.stats.lognorm.pdf:
                interval: the size interval of the distribution
                gsigma  : geometric standard deviation of distribution
                mode    : geometric mean radius of the particles
        """

        return np.array(
            [self.nparticles]
        ).sum()*self.pre_discretize()/self.volume


class ParticleInstances(ParticleDistribution):
    """ starting a particle distribution from single particles
    """

    def __init__(self, **kwargs):
        """  particle objects.
        """
        super().__init__(**kwargs)

        self.particle_radius = self.pre_radius() if kwargs.get(
            "particle_radius", None
        ) is None else in_radius(
            kwargs.get("particle_radius", None)
        )
        self.particle_number = self.nparticles if kwargs.get(
            "particle_radius", None
        ) is None else in_scalar(
            kwargs.get("particle_number", 1)
        )
        self.particle_density = in_density(
            kwargs.get("particle_density", 1000)
        )
        self.shape_factor = in_scalar(
            kwargs.get("shape_factor", 1)
        )
        self.volume_void = in_scalar(
            kwargs.get("volume_void", 0)
        )
        self.particle_charge = in_scalar(
            kwargs.get("particle_charge", 0)
        )
        self.particle_area_factor = in_scalar(
            kwargs.get("particle_area_factor", 1)
        )

    def particle_distribution(self):
        """ distribution
        """
        return (
            self.pre_distribution()
            if self.kwargs.get("particle_radius", None) is None
            else self.particle_number / self.particle_radius / self.volume
        )

    def particle_mass(self):
        """ Returns mass of particle.
        """
        return mass(
            radius=self.particle_radius,
            density=self.particle_density,
            shape_factor=self.shape_factor,
            volume_void=self.volume_void,
        )

    def particle_area(self):
        """ Returns particle's surface area
        """
        return area(
            radius=self.particle_radius,
            area_factor=self.particle_area_factor,
        )

    def knudsen_number(self):
        """ Returns particle's Knudsen number.
        """
        return knu(
            radius=self.particle_radius,
            mfp=self.mean_free_path(),
        )

    def slip_correction_factor(self):
        """ Returns particle's Cunningham slip correction factor.
        """
        return scf(
            radius=self.particle_radius,
            knu=self.knudsen_number(),
        )

    def friction_factor(self):
        """ Returns a particle's friction factor.
        """
        return frifac(
            radius=self.particle_radius,
            dynamic_viscosity=self.dynamic_viscosity(),
            scf=self.slip_correction_factor(),
        )


class ParticleCondensation(ParticleInstances):
    """ calculate some condensation stuff
    """

    def __init__(self, **kwargs):
        """ more particle objects.
        """
        super().__init__(**kwargs)
        self.kwargs = kwargs

    def molecular_enhancement(self):
        """ molecular enhancement
        """
        return mol_enh(
            vapor_size=self.vapor_radius,
            particle_size=self.particle_radius
        )

    def condensation_redmass(self):
        """ red mass
        """
        return redq(
            self.vapor_molec_wt,
            np.transpose([self.particle_mass().m]) *
            self.particle_mass().u*AVOGADRO_NUMBER
        ).squeeze()

    def vapor_speed(self):
        """ vapor speed
        """
        return cbar(
            temperature=self.temperature,
            molecular_weight=self.condensation_redmass(),
            gas_constant=self.gas_constant,
        )/4

    def fuchs_sutugin(self):
        """ the fuchs-sutugin correction
        """
        return fsc(
            knu_val=self.knudsen_number(), alpha=1
        )

    def vapor_flux(self):
        """ vapor flux
        """
        return phi(
            particle_area=self.particle_area(),
            molecular_enhancement=self.molecular_enhancement(),
            vapor_attachment=self.vapor_attachment,
            vapor_speed=self.vapor_speed(),
            driving_force=self.driving_force(),
            fsc=self.fuchs_sutugin(),
        )

    def particle_growth(self):
        """ particle growth in m/s
        """
        result = self.vapor_flux() * 2 / (
            self.vapor_density *
            np.transpose(
                [self.particle_radius.m**2]
                )*self.particle_radius.u**2 *
            np.pi *
            self.shape_factor
        )
        return (
            result if result.shape == self.particle_radius.shape
            else result.sum(axis=1)
        )


class Particle(ParticleCondensation):
    """ the Particle class!
    """

    def __init__(self, **kwargs):
        """ particle objects.
        """
        super().__init__(**kwargs)

        self.elementary_charge_value = in_handling(
            kwargs.get("elementary_charge_value", ELEMENTARY_CHARGE_VALUE),
            u.C
        )
        self.electric_permittivity = in_handling(
            kwargs.get("electric_permittivity", ELECTRIC_PERMITTIVITY),
            u.F/u.m
        )
        self.boltzmann_constant = in_handling(
            kwargs.get("boltzmann_constant", BOLTZMANN_CONSTANT),
            u.m**2*u.kg/u.s**2/u.K
        )
        self.coagulation_approximation = str(
            kwargs.get("coagulation_approximation", "hardsphere")
        )

        self.kwargs = kwargs

    def _coag_prep(self, other: "Particle"):
        """ get all related quantities to coulomb enhancement
        """
        return DimensionlessCoagulation(
            radius=self.particle_radius,
            other_radius=other.particle_radius,
            density=self.particle_density,
            other_density=other.particle_density,
            charge=self.particle_charge,
            other_charge=other.particle_charge,
            temperature=self.temperature,
            elementary_charge_value=self.elementary_charge_value,
            electric_permittivity=self.electric_permittivity,
            boltzmann_constant=self.boltzmann_constant,
            coag_approx=self.coagulation_approximation,
        )

    def reduced_mass(self, other: "Particle" = None):
        """ Returns the reduced mass.
        """
        return self._coag_prep(other or self).get_red_mass()

    def reduced_friction_factor(self, other: "Particle" = None):
        """ Returns the reduced friction factor between two particles.
        """
        return self._coag_prep(other or self).get_red_frifac()

    def coulomb_potential_ratio(self, other: "Particle" = None):
        """ Calculates the Coulomb potential ratio.
        """
        return self._coag_prep(other or self).get_ces()[0]

    def coulomb_enhancement_kinetic_limit(self, other: "Particle" = None):
        """ Kinetic limit of Coulomb enhancement for particle--particle cooagulation.
        """
        return self._coag_prep(other or self).get_ces()[1]

    def coulomb_enhancement_continuum_limit(self, other: "Particle" = None):
        """ Continuum limit of Coulomb enhancement for particle--particle coagulation.
        """
        return self._coag_prep(other or self).get_ces()[2]

    def diffusive_knudsen_number(self, other: "Particle" = None):
        """ Diffusive Knudsen number.
        """
        return self._coag_prep(other or self).get_diff_knu()

    def dimensionless_coagulation(self, other: "Particle" = None):
        """ Dimensionless particle--particle coagulation kernel.
        """
        return self._coag_prep(other or self).coag_less()

    def coagulation(self, other: "Particle" = None):
        """ Dimensioned particle--particle coagulation kernel
        """
        return self._coag_prep(other or self).coag_full()



""" discretization of the distribution of the particles
"""
import numpy as np
from scipy.stats import lognorm
from particula import u
from particula.util.input_handling import in_scalar, in_radius


def discretize(
    interval=None,
    disttype="lognormal",
    gsigma=in_scalar(1.25).m,
    mode=in_radius(100e-9).m,
    nparticles=in_scalar(1e5).m,
    **kwargs
):
    """ discretize the distribution of the particles

        Parameters:
            interval    (float) the size interval of the distribution
            distype     (str)   the type of distribution, "lognormal" for now
            gsigma      (float) geometric standard deviation of distribution
            mode        (float) pdf scale (corresponds to mode in lognormal)
    """

    _ = kwargs.get("something", None)

    if interval is None:
        raise ValueError("the 'interval' must be specified!")

    if not isinstance(interval, u.Quantity):
        interval = u.Quantity(interval, " ")

    if disttype != "lognormal":
        raise ValueError("the 'disttype' must be 'lognormal' for now!")

    return ((
        lognorm.pdf(
            x=interval.m,
            s=np.reshape(np.log(gsigma), (np.array([gsigma]).size, 1)),
            scale=np.reshape([mode], (np.array([mode]).size, 1)),
        ) / interval.u
        * np.reshape([nparticles], (np.array([nparticles]).size, 1))
        ).sum(axis=0) /
        np.array([nparticles]).sum() /
        np.max([np.array([mode]).size, np.array([gsigma]).size])
    )



""" a simple solver.

    For now, this takes as input:
    - the initial distribution (m**-4)
    - the radius of said distribution (m)
    - the coagulation kernel associated with it (m**3/s)
    - a desired time span in the form of ndarray (unitless, assumed in seconds)

    It returns an array of the with the dimensions
    (time_span, distribution)
    with the units of distribution (m**-4).
"""

import numpy as np
from scipy.integrate import odeint
from particula.util.coagulation_rate import CoagulationRate
from particula import u


def ode_func(_nums, _, _rads, _coag):
    """ function to integrate
    """
    coag = CoagulationRate(
        distribution=_nums * u.m**-4,
        radius=_rads * u.m,
        kernel=_coag * u.m**-4,
    )

    return coag.coag_gain().m - coag.coag_loss().m


class SimpleSolver:
    """ a class to solve the ODE:

        Need:
        1. initial distribution
        2. associated radius
        3. associated coagulation kernel

        Also:
        1. desired time span in seconds (given unitless)
    """

    def __init__(self, **kwargs):
        """ constructor

            kwargs:
            - distribution: initial distribution (m**-4)
            - radius: associated radius (m)
            - kernel: associated coagulation kernel (m**3/s)
            - tspan: desired time span (s)
        """
        self.nums_init = kwargs.get("distribution", None)
        self.rads_init = kwargs.get("radius", None)
        self.coag_kern = kwargs.get("kernel", None)
        self.time_span = kwargs.get("tspan", np.linspace(0, 10, 1000))
        self.kwargs = kwargs

    def prep_inputs(self):
        """ strip units, etc.
        """

        return (
            self.nums_init.m,
            self.rads_init.m,
            self.coag_kern.m,
            self.time_span,
        )

    def solution(self):
        """ utilize scipy.integrate.odeint
        """

        nums, rads, kern, time = self.prep_inputs()

        return odeint(
            ode_func,
            nums,
            time,
            args=(rads, kern)
        )*self.nums_init.u



""" stepping dynamically... one step at a time
"""


from particula.particle import Particle
from particula.util.coagulation_rate import CoagulationRate
from particula.util.dimensionless_coagulation import full_coag


class DynamicStep(Particle):
    """ The class to calculate the rate at one step
    """

    def __init__(self, **kwargs):
        """ Constructiong the class and inheriting from the
            ParticleDistribution class.

            See the ParticleDistribution class for the parameters:
            under particula.aerosol_dynamics.particle_distribution
        """

        super().__init__(**kwargs)

        self.kwargs = kwargs

    def coag_kern(self):
        """ get coagulation kernel

            see:
            particula.util.dimensionless_coagulation:
            function full_coag
        """

        return full_coag(radius=self.particle_radius, **self.kwargs)

    def coag_loss(self):
        """ get the coagulation loss rate from:

            particula.util.coagulation_rate:
            class CoagulationRate
            function coag_loss
        """

        return CoagulationRate(
            distribution=self.particle_distribution(),
            radius=self.particle_radius,
            kernel=self.coag_kern(),
        ).coag_loss()

    def coag_gain(self):
        """ get coagulation gain rate from:

            particula.util.coagulation_rate:
            class CoagulationRate
            function coag_gain
        """

        return CoagulationRate(
            distribution=self.particle_distribution(),
            radius=self.particle_radius,
            kernel=self.coag_kern(),
        ).coag_gain()

    def coag_rate(self):
        """ get the coagulation rate by summing the loss and gain rates
        """

        return self.coag_gain() - self.coag_loss()



""" calculating the mean free path of air

    The mean free path is the average distance
    traveled by a molecule between collisions
    with other molecules present in a medium (air).

    The expeected mean free path of air is approx.
    65 nm at 298 K and 101325 Pa.

    TODO:
        add size checks for pressure--temperature pairs
        to ensure that they match; otherwise, an error will occur
        or use broadcast (though this is likely not a good idea)?
        perhaps allow for a height--temperature--pressure dependency
        somewhere? this could be import for @Gorkowski's parcels...
        (likely through a different utility function...)
"""

import numpy as np
from particula import u
from particula.constants import GAS_CONSTANT, MOLECULAR_WEIGHT_AIR
from particula.util.dynamic_viscosity import dyn_vis
from particula.util.input_handling import (in_gas_constant,
                                           in_molecular_weight, in_pressure,
                                           in_temperature, in_viscosity)


def mfp(
    temperature=298.15*u.K,
    pressure=101325*u.Pa,
    molecular_weight=MOLECULAR_WEIGHT_AIR,
    dynamic_viscosity=None,
    gas_constant=GAS_CONSTANT,
    **kwargs,
):
    """ Returns the mean free path of in air.

        The mean free path is the average distance
        traveled by a molecule between collisions
        with other molecules present in a medium (air).

        The expeected mean free path of air is approx.
        65 nm at 298 K and 101325 Pa.

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.mean_free_path import mfp
        >>> # with no inputs, it defaults to 298 K and 101325 Pa
        >>> mfp()
        <Quantity(6.64373669e-08, 'meter')>
        >>> # specifying a temperature of 300 K
        >>> mfp(temperature=300*u.K).magnitude
        6.700400687925813e-08
        >>> # specifying 300 K and  pressure of 1e5 Pa
        >>> mfp(temperature=300*u.K, pressure=1e5*u.Pa)
        <Quantity(6.789181e-08, 'meter')>
        >>> mfp(
        ... temperature=300,
        ... pressure=1e5,
        ... molecular_weight=0.03
        ... )
        <Quantity(6.67097062e-08, 'meter')>
        >>> # specifying explicit value for dynamic viscosity
        >>> mfp(dynamic_viscosity=1e-5)
        <Quantity(3.61864151e-08, 'meter')>
        >>> # specifying implicit value for dynamic viscosity
        >>> mfp(
        ... temperature=300,
        ... reference_viscosity=1e-5,
        ... reference_temperature=273.15
        ... )
        <Quantity(3.90466241e-08, 'meter')>
        >>> # specifying a list of temperatures
        >>> mfp(temperature=[200, 250, 300, 400]).m
        array([3.93734886e-08, 5.29859655e-08, 6.70040069e-08, 9.57800224e-08])
        >>> # specifying a list of pressures
        >>> mfp(pressure=[1.0e5, 1.1e5, 1.2e5, 1.3e5]).m
        array([6.73607078e-08, 6.12370071e-08, 5.61339232e-08, 5.18159291e-08])
        >>> # specifying a list of pressures and temperatures
        >>> mfp(temperature=[300,310], pressure=[1e5, 1.1e5])
        <Quantity([6.78918100e-08 6.43354325e-08], 'meter')>
        ```

        Parameters: (either # or $)
            temperature           (float) [K]      (default: 298)
            pressure              (float) [Pa]     (default: 101325)
            molecular_weight      (float) [kg/mol] (default: constants)

        #   dynamic_viscosity     (float) [Pa*s]   (default: util)
        $   reference_viscosity   (float) [Pa*s]   (default: constants)
        $   reference_temperature (float) [K]      (default: constants)

        Returns:
                        (float) [m]

        Using particula.constants:
            GAS_CONSTANT            (float) [J/mol/K]
            MOLECULAR_WEIGHT_AIR    (float) [kg/mol]

            REF_VISCOSITY_AIR_STP   (float) [Pa*s]
            REF_TEMPERATURE_STP     (float) [K]
            SUTHERLAND_CONSTANT     (float) [K]

        Notes:
            dynamic_viscosity can be calculated independently via
            particula.util.dynamic_viscosity.dyn_vis(**kwargs), but
            if the value of dynamic_viscosity is provided directly,
            it overrides the calculated value.
    """

    temp = in_temperature(temperature)
    pres = in_pressure(pressure)
    molec_wt = in_molecular_weight(molecular_weight)

    if dynamic_viscosity is None:
        dyn_vis_val = in_viscosity(
            dyn_vis(temperature=temp, **kwargs)
        )
    else:
        dyn_vis_val = in_viscosity(dynamic_viscosity)

    gas_con = in_gas_constant(gas_constant)

    return (
        (2 * dyn_vis_val / pres) /
        (8 * molec_wt / (np.pi * gas_con * temp))**0.5
    ).to_base_units()



""" Calculate the coagulation rate.
"""

import numpy as np
from scipy.interpolate import RectBivariateSpline


class CoagulationRate:
    """ A class to calculate the coagulation rate.

        Parameters:
            distribution (float): The distribution of particles.
            radius       (float): The radius of the particles.
            kernel       (float): The kernel of the particles.
    """

    def __init__(self, distribution, radius, kernel):
        """ Constructing the class, needing already built
            distribution, radius and kernel.

            * the distribution has units of m**-4
            * the radius has units of m
            * the kernel has units of m**3/s

            Note: we strip the units first because we want
                  to allow solvers to use this without issue.
                  These solvers often need pure ndarrays.
                  It would work fine with the units, but
                  it will be throwing warning about stripping
                  them, which is annoying and maybe not good
                  for performance.
        """

        self.distribution = distribution
        self.radius = radius
        self.kernel = kernel

    def coag_prep(self):
        """ Repackage the parameters
        """

        nums = self.distribution
        rads = self.radius
        kern = self.kernel

        return nums, rads, kern

    def coag_loss(self):
        """ Returns the coagulation loss rate

            Equation:

            loss_rate(other_radius) = (
                dist(other_radius) *
                integral( # over all space
                    kernel(radius, other_radius) *
                    dist(radius),
                    radius
                )

            Units:

            m**-4/s
        """

        nums, rads, kern = self.coag_prep()

        return nums*np.trapz(kern*nums, rads)

    def coag_gain(self):
        """ Returns the coagulation gain rate

            Equation:

            gain_rate(other_radius) = (
                other_radius**2 *
                integral( # from some_radius=0 to other_radius/2**(1/3)
                    kernel(some_radius, (other_radius**3-some_radius**3)*(1/3)*
                    dist(some_radius) *
                    dist((other_radius**3 - some_radius**3)*(1/3)) /
                    (other_radius**3 - some_radius**3)*(2/3),
                    some_radius
                )
            )

            Units:
                m**-4/s

            This equation necessitates the use of a for-loop due to the
            convoluted use of different radii at different stages.
            This is the costliest step of all coagulation calculations.
            Note, to estimate the kernel and distribution at
            (other_radius**3 - some_radius**3)*(1/3)
            we use interporlation techniques.

            Using `RectBivariateSpline` accelerates this significantly.
        """

        nums, rads, kern = self.coag_prep()

        interp = RectBivariateSpline(
            rads.m, rads.m, kern.m * nums.m * np.transpose([nums.m])
        )

        dpd = np.linspace(0, rads.m/2**(1/3), rads.m.size)*rads.u
        dpi = ((np.transpose(rads.m)*rads.u)**3 - dpd**3)**(1/3)

        return rads**2 * np.trapz(
            interp.ev(dpd.m, dpi.m) * kern.u * nums.u * nums.u / dpi**2,
            dpd,
            axis=0
        )



""" calculating rms speed of molecules
"""

import numpy as np
from particula.constants import GAS_CONSTANT, MOLECULAR_WEIGHT_AIR
from particula.util.input_handling import (in_gas_constant,
                                           in_molecular_weight, in_temperature)


def cbar(
    temperature=298.15,
    molecular_weight=MOLECULAR_WEIGHT_AIR,
    gas_constant=GAS_CONSTANT,
):
    """ Returns the mean speed of molecules in an ideal gas.

        Parameters:
            temperature           (float) [K]      (default: 298.15)
            molecular_weight      (float) [kg/mol] (default: constants)

        Returns:
                                  (float) [m/s]

        Using particula.constants:
            GAS_CONSTANT            (float) [J/mol/K]
            MOLECULAR_WEIGHT_AIR    (float) [kg/mol]

    """

    temperature = in_temperature(temperature)
    molecular_weight = in_molecular_weight(molecular_weight)
    gas_constant = in_gas_constant(gas_constant)

    return (
        (8 * gas_constant * temperature/(np.pi * molecular_weight))**0.5
    ).to_base_units()



""" calculate coulombic enhancements

    TODO:
        * add clarification about charge values (scalar or array)
        * for now, assume scalar charge values only (i.e. one charge at a time)
        * explain ratio.all() or ratio.any() control flow
"""

import numpy as np
from particula import u
from particula.constants import (BOLTZMANN_CONSTANT, ELECTRIC_PERMITTIVITY,
                                 ELEMENTARY_CHARGE_VALUE)
from particula.util.input_handling import (in_handling, in_radius, in_scalar,
                                           in_temperature)


class CoulombEnhancement:  # pylint: disable=too-many-instance-attributes

    """ Calculate coulomb-related enhancements

        Attributes:
            radius          (float) [m]
            other_radius    (float) [m]             (default: radius)
            charge          (float) [dimensionless] (default: 0)
            other_charge    (float) [dimensionless] (default: 0)
            temperature     (float) [K]             (default: 298)
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        radius=None,
        other_radius=None,
        charge=0,
        other_charge=0,
        temperature=298,
        elementary_charge_value=ELEMENTARY_CHARGE_VALUE,
        electric_permittivity=ELECTRIC_PERMITTIVITY,
        boltzmann_constant=BOLTZMANN_CONSTANT,
        **kwargs
    ):
        """ Initialize CoulombEnhancement object
        """

        other_radius = radius if other_radius is None else other_radius

        self.radius = in_radius(radius)
        self.other_radius = in_radius(other_radius)
        self.charge = in_scalar(charge)
        self.other_charge = in_scalar(other_charge)
        self.temperature = in_temperature(temperature)
        self.elem_char_val = in_handling(elementary_charge_value, u.C)
        self.elec_perm = in_handling(electric_permittivity, u.F/u.m)
        self.boltz_const = in_handling(
            boltzmann_constant, u.m**2*u.kg/u.s**2/u.K
        )

        self.kwargs = kwargs

        if np.array(self.temperature.m).size > 1:
            raise ValueError(
                f"\t\n"
                f"\tTemperature {self.temperature} must be scalar for this!\n"
                f"\tThis is to allow calculation of the coagulation kernel.\n"
                f"\tYou can repeat this routine for different temperatures."
            )

        if np.array(
            self.charge.m
        ).size + np.array(self.other_charge.m).size > 2:
            raise ValueError(
                f"\t\n"
                f"\t Charges {self.charge} and {self.other_charge} must\n"
                f"\tbe scalars for this to work!\n"
                f"\tYou can repeat this routine for different charges."
            )

    def coulomb_potential_ratio(self):
        """ Calculates the Coulomb potential ratio.

            Parameters:
                radius          (float) [m]
                other_radius    (float) [m]             (default: radius)
                charge          (int)   [dimensionless] (default: 0)
                other_charge    (int)   [dimensionless] (default: 0)
                temperature     (float) [K]             (default: 298)

            Returns:
                                (float) [dimensionless]
        """

        numerator = -1 * self.charge * self.other_charge * (
            self.elem_char_val ** 2
        )
        denominator = 4 * np.pi * self.elec_perm * (
            np.transpose([self.radius.m])*self.radius.u +
            self.other_radius
        )

        return numerator / (
            denominator * self.boltz_const * self.temperature
        )

    def coulomb_enhancement_kinetic_limit(self):
        """ Coulombic coagulation enhancement kinetic limit.

            Parameters:
                radius          (float) [m]
                other_radius    (float) [m]             (default: radius)
                charge          (float) [dimensionless] (default: 0)
                other_charge    (float) [dimensionless] (default: 0)
                temperature     (float) [K]             (default: 298)

            Returns:
                                (float) [dimensionless]
        """

        ratio = self.coulomb_potential_ratio()

        return (
            (1 + ratio) if ratio.all() >= 0
            else np.exp(ratio)
        ).to_base_units()

    def coulomb_enhancement_continuum_limit(self):
        """ Coulombic coagulation enhancement continuum limit.

            Parameters:
                radius          (float) [m]
                other_radius    (float) [m]             (default: radius)
                charge          (float) [dimensionless] (default: 0)
                other_charge    (float) [dimensionless] (default: 0)
                temperature     (float) [K]             (default: 298)

            Returns:
                                (float) [dimensionless]
        """

        ratio = self.coulomb_potential_ratio()

        return (
            ratio/(1-np.exp(-1*ratio)) if ratio.all() != 0
            else u.Quantity(1, u.dimensionless)
        ).to_base_units()


def cpr(**kwargs):
    """ Calculate coulomb potential ratio
    """
    return CoulombEnhancement(**kwargs).coulomb_potential_ratio()


def cekl(**kwargs):
    """ Calculate coulombic enhancement kinetic limit
    """
    return CoulombEnhancement(**kwargs).coulomb_enhancement_kinetic_limit()


def cecl(**kwargs):
    """ Calculate coulombic enhancement continuum limit
    """
    return CoulombEnhancement(**kwargs).coulomb_enhancement_continuum_limit()


def coulomb_enhancement_all(**kwargs):
    """ Return all the values above in one call
    """
    cea = CoulombEnhancement(**kwargs)
    return (
        cea.coulomb_potential_ratio(),
        cea.coulomb_enhancement_kinetic_limit(),
        cea.coulomb_enhancement_continuum_limit()
    )



""" calculate slip correction
"""

import numpy as np
from particula.util.input_handling import in_radius, in_scalar
from particula.util.knudsen_number import knu as knu_func


def scf(
    radius=None,
    knu=None,
    **kwargs
):
    """ Returns particle's Cunningham slip correction factor.

        Dimensionless quantity accounting for non-continuum effects
        on small particles. It is a deviation from Stokes' Law.
        Stokes assumes a no-slip condition that is not correct at
        high Knudsen numbers. The slip correction factor is used to
        calculate the friction factor.

        Thus, the slip correction factor is about unity (1) for larger
        particles (Kn -> 0). Its behavior on the other end of the
        spectrum (smaller particles; Kn -> inf) is more nuanced, though
        it tends to scale linearly on a log-log scale, log Cc vs log Kn.

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.slip_correction import scf
        >>> # with radius 1e-9 m
        >>> scf(radius=1e-9)
        <Quantity(110.720731, 'dimensionless')>
        >>> # with radius 1e-9 m and knu=1
        >>> scf(radius=1e-9*u.m, knu=1)
        <Quantity(2.39014843, 'dimensionless')>
        >>> # using knu(**kwargs)
        >>> scf(radius=1e-9*u.m, mfp=60*u.nm)
        <Quantity(99.9840088, 'dimensionless')>
        >>> # using mfp(**kwargs)
        >>> scf(
        ... radius=1e-9*u.m,
        ... temperature=300,
        ... pressure=1e5,
        ... molecular_weight=0.03
        ... )
        <Quantity(111.101591, 'dimensionless')>
        ```

        Parameters:
            radius  (float) [m]
            knu     (float) [ ] (default: util)

        Returns:
                    (float) [dimensionless]

        Notes:
            knu can be calculated using knu(**kwargs);
            refer to particula.util.knudsen_number.knu for more info.

    """
    radius = in_radius(radius)

    knu_val = knu_func(radius=radius, **kwargs) if knu is None \
        else in_scalar(knu)

    return 1 + knu_val * (
        1.257 + 0.4*np.exp(-1.1/knu_val)
    )



""" Calculate the diffusive knudsen number
"""

import numpy as np
from particula.constants import BOLTZMANN_CONSTANT
from particula.util.coulomb_enhancement import (CoulombEnhancement,
                                                coulomb_enhancement_all)
from particula.util.friction_factor import frifac
from particula.util.input_handling import in_density
from particula.util.particle_mass import mass
from particula.util.reduced_quantity import reduced_quantity


class DiffusiveKnudsen(CoulombEnhancement):
    """ A class for Diff..Knu
    """

    def __init__(
        self,
        density=1000,
        other_density=None,
        **kwargs
    ):
        """ define properties """

        super().__init__(**kwargs)

        self.density = in_density(density)
        self.other_density = self.density if other_density is None \
            else in_density(other_density)

        self.kwargs = kwargs

    def get_red_mass(self):
        """ get the reduced mass
        """

        mass_dummy = mass(radius=self.radius, density=self.density)

        return reduced_quantity(
            np.transpose([mass_dummy.m])*mass_dummy.u,
            mass(radius=self.other_radius, density=self.other_density)
        )

    def get_rxr(self):
        """ add two radii
        """
        return np.transpose([self.radius.m])*self.radius.u + self.other_radius

    def get_red_frifac(self):
        """ get the reduced friction factor
        """

        frifac_kwargs = self.kwargs.copy()
        other_frifac_kwargs = self.kwargs.copy()

        frifac_kwargs.pop("radius", None)
        other_frifac_kwargs.pop("radius", None)

        dummy_frifac = frifac(radius=self.radius, **frifac_kwargs)
        return reduced_quantity(
            np.transpose([dummy_frifac.m])*dummy_frifac.u,
            frifac(radius=self.other_radius, **other_frifac_kwargs)
        )

    def get_ces(self):
        """ get coulomb enhancement parameters
        """
        lkwargs = self.kwargs.copy()
        return coulomb_enhancement_all(**lkwargs)

    def get_celimits(self):
        """ get coag enh limits
        """

        lkwargs = self.kwargs.copy()
        return coulomb_enhancement_all(**lkwargs)[1:]

    def get_diff_knu(self):
        """ calculate it
        """
        boltz_const = BOLTZMANN_CONSTANT
        rmass = self.get_red_mass()
        rfrifac = self.get_red_frifac()
        (cekl_val, cecl_val) = self.get_celimits()
        temp = self.temperature
        return (
            ((temp * boltz_const * rmass)**0.5 / rfrifac) /
            (self.get_rxr() * cecl_val / cekl_val)
        )


def diff_knu(**kwargs):
    """ Diffusive Knudsen number.

        The *diffusive* Knudsen number is different from Knudsen number.
        Ratio of:
            - numerator: mean persistence of one particle
            - denominator: effective length scale of
                particle--particle Coulombic interaction

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.diffusive_knudsen import diff_knu
        >>> # with only one radius
        >>> diff_knu(radius=1e-9)
        <Quantity(29.6799, 'dimensionless')>
        >>> # with two radii
        >>> diff_knu(radius=1e-9, other_radius=1e-8)
        <Quantity(3.85387845, 'dimensionless')>
        >>> # with radii and charges
        >>> diff_knu(radius=1e-9, other_radius=1e-8, charge=-1, other_charge=1)
        <Quantity(4.58204028, 'dimensionless')>
        ```
        Parameters:
            radius          (float) [m]
            other_radius    (float) [m]             (default: radius)
            density         (float) [kg/m^3]        (default: 1000)
            other_density   (float) [kg/m^3]        (default: density)
            charge          (int)   [dimensionless] (default: 0)
            other_charge    (int)   [dimensionless] (default: 0)
            temperature     (float) [K]             (default: 298)

        Returns:
                            (float) [dimensionless]

        Notes:
            this function uses the friction factor and
            the coulomb enhancement calculations; for more information,
            please see the documentation of the respective functions:
                - particula.util.friction_factor.frifac(**kwargs)
                - particula.util.coulomb_enhancement.cekl(**kwargs)
                - particula.util.coulomb_enhancement.cecl(**kwargs)

    """

    return DiffusiveKnudsen(**kwargs).get_diff_knu()


def red_mass(**kwargs):
    """ get the reduced mass
    """

    return DiffusiveKnudsen(**kwargs).get_red_mass()


def red_frifac(**kwargs):
    """ get the reduced friction factor
    """

    return DiffusiveKnudsen(**kwargs).get_red_frifac()


def ces(**kwargs):
    """ get the coulomb enhancement limits
    """

    return DiffusiveKnudsen(**kwargs).get_ces()


def celimits(**kwargs):
    """ get coag enh limits
    """

    return DiffusiveKnudsen(**kwargs).get_celimits()


def rxr(**kwargs):
    """ add two radii
    """

    return DiffusiveKnudsen(**kwargs).get_rxr()



""" calculating reduced quantity.

    reduced_quantity =
        quantity_1 * quantity_2 / (quantity_1 + quantity_2)
"""

from particula import u


def reduced_quantity(a_quantity, b_quantity):
    """ Returns the reduced mass of two particles.

        Examples:
        ```
        >>> reduced_quantity(1*u.kg, 1*u.kg)
        <Quantity(0.5, 'kilogram')>
        >>> reduced_quantity(1*u.kg, 20*u.kg).m
        0.9523809523809523
        >>> reduced_quantity(1, 200)
        0.9950248756218906
        >>> reduced_quantity([1, 2, 3], 200)
        array([0.99502488, 1.98019802, 2.95566502])
        >>> reduced_quantity([1, 2], [200, 300])
        array([0.99502488, 1.98675497])
        ```

        Parameters:
            a_quantity  (float)  [arbitrary units]
            b_quantity  (float)  [arbitrary units]

        Returns:
                        (float)  [arbitrary units]

        A reduced quantity is an "effective inertial" quantity,
        allowing two-body problems to be solved as one-body problems.
    """

    a_q = a_quantity
    b_q = b_quantity

    if isinstance(a_q, u.Quantity):
        a_q = a_q.to_base_units()
        if not isinstance(b_q, u.Quantity):
            raise TypeError(
                f"\n\t"
                f"{a_q} and {b_q} (dimensionless) not compatible!\n\t"
                f"Quantities must have same units to be reduced.\n\t"
                f"Try: {a_q} and {b_q} {a_q.units} for example.\n"
            )
        if a_q.units != b_q.to_base_units().units:
            raise TypeError(
                f"\n\t"
                f"{a_q} and {b_q} not compatible!\n"
                f"Quantities must have same units to be reduced.\n\t"
                f"Try: {a_q} and {b_q} {a_q.units} for example."
            )
    elif isinstance(b_q, u.Quantity):
        b_q = b_q.to_base_units()
        if not isinstance(a_q, u.Quantity):
            raise TypeError(
                f"\n\t"
                f"{a_q} (dimensionless) and {b_q} not compatible!\n\t"
                f"Quantities must have same units to be reduced.\n\t"
                f"Try: {b_q} and {a_q} {b_q.units} for example."

            )
        if a_q.to_base_units().units != b_q.units:
            raise TypeError(
                f"\n\t"
                f"{a_q} and {b_q} not compatible!\n\t"
                f"Quantities must have same units to be reduced.\n\t"
                f"Try: {b_q} and {a_q} {b_q.units} for example"
            )

    if not isinstance(a_q, u.Quantity):
        a_q = u.Quantity(a_q, " ")
    if not isinstance(b_q, u.Quantity):
        b_q = u.Quantity(b_q, " ")

    return (a_q * b_q / (a_q + b_q)).to_base_units()



""" A class with methods for dimensionless coagulation
"""

from particula.util.input_handling import in_scalar
from particula.util.diffusive_knudsen import DiffusiveKnudsen as DKn
from particula.util.diffusive_knudsen import celimits
# from particula.util.diffusive_knudsen import diff_knu as dknu
from particula.util.diffusive_knudsen import red_frifac, red_mass, rxr
from particula.util.approx_coagulation import approx_coag_less


class DimensionlessCoagulation(DKn):

    """ dimensionless coagulation
    """

    def __init__(
        self,
        dkn_val=None,
        coag_approx="hardsphere",
        **kwargs
    ):
        """ Dimensionless particle--particle coagulation kernel.

            Attributes:
                diff_knu        (float) [dimensionless]

            Notes:
                The dimensionless coagulation kernel is defined as
                a function of the diffusive knudsen number; for more info,
                please see the documentation of the respective function:
                    - particula.util.diffusive_knudsen.diff_knu(**kwargs)
        """
        super().__init__(**kwargs)

        self.diff_knu = DKn(**kwargs).get_diff_knu() if dkn_val is None \
            else in_scalar(dkn_val)

        self.coag_approx = coag_approx

        self.kwargs = kwargs

    def coag_less(self):
        """ Return the dimensionless coagulation kernel.
        """

        impls = ["hardsphere", "gh2012", "cg2019", "dy2007", "gk2008"]

        if self.coag_approx not in impls:
            raise ValueError(f"{self.coag_approx} not recognized!")

        return approx_coag_less(
            diff_knu=self.diff_knu,
            cpr=self.coulomb_potential_ratio(),
            approx=self.coag_approx
        )

    def coag_full(self):
        """ Retrun the dimensioned coagulation kernel
        """

        coag = self.coag_less()
        redff = red_frifac(**self.kwargs)
        redm = red_mass(**self.kwargs)
        cekl, cecl = celimits(**self.kwargs)
        xrxr = rxr(**self.kwargs)

        return (
            coag * redff * xrxr**3 * cekl**2 / (redm * cecl)
        )


def less_coag(**kwargs):
    """ Return the dimensionless coagulation kernel.

        The dimensionless coagulation kernel is defined as
        a function of the diffusive knudsen number; for more info,
        please see the documentation of the respective function:
            - particula.util.diffusive_knudsen.diff_knu(**kwargs)

        Examples:
        ```
        >>> from particula import u
        >>> from particula.util.dimensionless_coagulation import less_coag
        >>> # only for hardsphere coagulation for now
        >>> # with only one radius
        >>> less_coag(radius=1e-9)
        <Quantity(147.877572, 'dimensionless')>
        >>> # with two radii
        >>> less_coag(radius=1e-9, other_radius=1e-8)
        <Quantity(18.4245966, 'dimensionless')>
        >>> # with two radii and charges
        >>> less_coag(
        ... radius=1e-9, other_radius=1e-8, charge=1, other_charge=-1
        ... )
        <Quantity(22.0727435, 'dimensionless')>
    """

    return DimensionlessCoagulation(**kwargs).coag_less()


def full_coag(**kwargs):
    """ Return the dimensioned coagulation kernel
    """

    return DimensionlessCoagulation(**kwargs).coag_full()
